(** Like the logic module [Atom] plus the possibility to order atoms in a taxonomy.

   The taxonomy links are handled by the module [Term].

*)

module type PARAM = sig
  val names : string list (** The list of admissible names for atoms. If this list is empty, every name is admissible (open set of atoms). *)
end

module Make (Param : PARAM) : Logic.T =
  struct
    include Logic.Default

    type t = Attr of string | Term of string

    let isvalue = function
      | Attr _ -> true
      | Term t -> false (*Term.get_sufs t = []*)

    let parse = parser
      |	[<'Token.Term n
	  when not (List.mem n Syntax.keywords)>] -> Term n
      |	[<'Token.Ident a
	  when (Param.names = [] or List.mem a Param.names)
	      & not (List.mem a Syntax.keywords)>] -> Attr a

    let parse_compact = parser
      | [<'Token.String s>] ->
          let res = ref [] in
          let l = String.length s in
          for i = l - 1 downto 0 do
            match s.[i] with
            | 'a' .. 'z' -> res := Token.Ident (String.sub s i 1) :: !res
            | 'A' .. 'Z' -> res := Token.Term (String.sub s i 1) :: !res
            | ' ' | '\t' | '\n' -> res := Token.Term "_" :: !res
            | _ -> ()
          done;
          !res

    let print = function
      | Attr a -> [Token.Ident a]
      |	Term name -> [Token.Term name]

    let mpt l = List.map (fun toks -> parse (Syntax.of_list toks)) l

    let top () = raise Not_found

    let rec entails a b = match a,b with
    | Attr x, Attr y -> x = y
    | Attr _, Term n -> List.exists (fun b' -> entails a b') (mpt (Term.get_sufs n))
    | Term m, Attr _ -> List.exists (fun a' -> entails a' b) (mpt (Term.get_necs m))
    | Term m, Term n -> m=n or
	List.exists (fun b' -> entails a b') (mpt (Term.get_sufs n)) or
	List.exists (fun a' -> entails a' b) (mpt (Term.get_necs m))

    let conj a b =
      if entails a b then a 
      else if entails b a then b
      else raise Not_found

    let add a b = b

    let sub a b = if entails a b then raise Not_found else a

    let rec features = function
      | Attr a -> [true,Attr a]
      |	Term n -> List.fold_left (fun res c -> LSet.union res (features c)) [true,Term n] (mpt (Term.get_necs n))

    let axiom a b_opt =
      match a, b_opt with
      | Term x, None ->
	  List.iter
	    (fun nec -> Term.del_nec x nec)
	    (Term.get_necs x);
	  LSet.singleton x
      | Term x, Some (Term y) ->
	  if entails a (Term y)
	  then begin
	    List.iter
	      (fun nec -> Term.del_nec x nec)
	      (List.filter
		 (function
		   | [Token.Term z] -> entails (Term z) (Term y)
		   | _ -> false)
		 (Term.get_necs x));
	    Term.add_nec x [Token.Term y] end
	  else begin
	    List.iter
	      (fun nec -> Term.del_nec x nec)
	      (List.filter
		 (function
		   | [Token.Term z] -> entails (Term y) (Term z)
		   | _ -> false)
		 (Term.get_necs x));
	    Term.add_nec x [Token.Term y] end;
	  LSet.singleton x
      | _, _ -> LSet.empty ()

    let rec gen y d gs =
      List.fold_left
        (fun res x ->
           if List.exists (fun g -> entails g x) res
           then res
           else LSet.add x (List.filter (fun g -> not (entails x g)) res))
        gs
        (LSet.inter (List.map snd (features y)) (List.map snd (features d)))
(*
      List.filter
        (fun x -> not (List.exists (fun g -> entails g x) gs))
        (LSet.inter (features y) (features d))
*)

(*
    let rec gen2 b c al =
        let xs = List.filter
          (fun x -> not (List.exists (fun a -> entails a x) al))
          ( match b with
          | Attr _ -> []
          | Term n -> mpt (Term.get_necs n)) in
        List.fold_right
          (fun x res ->
             if List.mem x res then res
             else if entails c x then LSet.add x res
             else LSet.union res (gen2 x c al))
          xs
          []
    and gen b c al =
      if entails c b then [b]
      else if entails b c then [c]
      else gen2 b c al
*)
  end

