(** Dates at 3 levels of granularity: years, months, and days. *)

module Make : Logic.T =
  struct
    include Logic.Default

    type t = (int option * int) option * int  (* day, month, year *)

    open Unix

    let from_time : float -> t =
      fun sec ->
	let tm = localtime sec in
	(Some (Some tm.tm_mday, tm.tm_mon+1), tm.tm_year+1900)

    let day : float = 86400.
    let today : unit -> t = fun () -> from_time (time ())
    let yesterday : unit -> t = fun () -> from_time (time () -. day)
    let tomorrow : unit -> t = fun () -> from_time (time () +. day)

    let this_month : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (Some (None, tm.tm_mon+1), tm.tm_year+1900)
    let next_month : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (Some (None, ((tm.tm_mon + 1) mod 12) + 1), tm.tm_year + 1900 + if tm.tm_mon=11 then 1 else 0)
    let last_month : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (Some (None, ((tm.tm_mon - 1) mod 12) + 1), tm.tm_year + 1900 - if tm.tm_mon=0 then 1 else 0)

    let this_year : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, tm.tm_year+1900)
    let next_year : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, tm.tm_year+1901)
    let last_year : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, tm.tm_year+1899)


    let isvalue = function
      |	Some (Some d, m), y -> true
      | _ -> false

    let compare (dm1,y1) (dm2,y2) =
      if y1 < y2 then -1
      else if y1 > y2 then 1
      else match dm1, dm2 with
      |	None, _
      |	_, None -> 0
      |	Some (d1,m1), Some (d2,m2) ->
	  if m1 < m2 then -1
	  else if m1 > m2 then 1
	  else match d1, d2 with
	  | None, _
	  | _, None -> 0
	  | Some n1, Some n2 ->
	      if n1 < n2 then -1
	      else if n1 > n2 then 1
	      else 0

    let entails (dm1,y1) (dm2,y2) =
      y1 = y2 &
      match dm1, dm2 with
      |	_, None -> true
      |	Some (d1,m1), Some (d2,m2) ->
	  m1 = m2 &
	  ( match d1, d2 with
	  | _, None -> true
	  | Some n1, Some n2 -> n1 = n2
	  | _ -> false)
      |	_ -> false

    let conj d1 d2 =
      if entails d1 d2 then d1
      else if entails d2 d1 then d2
      else raise Not_found

    let add d1 d2 = conj d1 d2

    let sub d1 d2 = if entails d1 d2 then raise Not_found else d1

    let features = function
      |	None, y -> [true,(None, y)]
      |	Some (None, m), y -> [true,(Some (None, m), y); true,(None,y)]
      |	Some (Some d, m), y -> [true,(Some (Some d, m), y); true,(Some (None, m), y); true,(None,y)]


    open Token

    let parse_month = parser
      |	[<'Ident "jan">] -> 1
      |	[<'Ident "feb">] -> 2
      |	[<'Ident "mar">] -> 3
      |	[<'Ident "apr">] -> 4
      |	[<'Ident "may">] -> 5
      |	[<'Ident "jun">] -> 6
      |	[<'Ident "jul">] -> 7
      |	[<'Ident "aug">] -> 8
      |	[<'Ident "sep">] -> 9
      |	[<'Ident "oct">] -> 10
      |	[<'Ident "nov">] -> 11
      |	[<'Ident "dec">] -> 12

    let print_month = function
      |	1 -> Ident "jan"
      |	2 -> Ident "feb"
      |	3 -> Ident "mar"
      |	4 -> Ident "apr"
      |	5 -> Ident "may"
      |	6 -> Ident "jun"
      |	7 -> Ident "jul"
      |	8 -> Ident "aug"
      |	9 -> Ident "sep"
      |	10 -> Ident "oct"
      |	11 -> Ident "nov"
      |	_ -> Ident "dec"

    let parse = parser
      |	[<'Int dy; s>] ->
	  ( match s with parser
	  | [<m = parse_month; 'Int y ?? "Syntax error: year expected after: " ^ Syntax.stringizer [Int dy; print_month m]>] ->
	      (Some (Some dy, m), y)
	  | [<>] ->
	      (None, dy)
	  )
      |	[<m = parse_month; 'Int y ?? "Syntax error: year expected after: " ^ Syntax.stringizer [print_month m]>] ->
	  (Some (None, m), y)
      |	[<'Ident "today">] -> today ()
      |	[<'Ident "tomorrow">] -> tomorrow ()
      |	[<'Ident "yesterday">] -> yesterday ()
      |	[<'Ident "this"; str>] -> begin match str with parser
	| [<'Ident "day">] -> today ()
	| [<'Ident "month">] -> this_month ()
	| [<'Ident "year">] -> this_year ()
	| [<>] -> raise (Stream.Error "Syntax error: 'day', 'month' or 'year' expected after 'this'") end
      |	[<'Ident "next"; str>] -> begin match str with parser
	| [<'Ident "day">] -> tomorrow ()
	| [<'Ident "month">] -> next_month ()
	| [<'Ident "year">] -> next_year ()
	| [<>] -> raise (Stream.Error "Syntax error: 'day', 'month' or 'year' expected after 'next'") end
      |	[<'Ident "last"; str>] -> begin match str with parser
	| [<'Ident "day">] -> yesterday ()
	| [<'Ident "month">] -> last_month ()
	| [<'Ident "year">] -> last_year ()
	| [<>] -> raise (Stream.Error "Syntax error: 'day', 'month' or 'year' expected after 'last'") end

    let rec print = function
      |	None, y -> [Int y]
      |	Some (None, m), y -> [print_month m; PP_tilda; Int y]
      |	Some (Some d, m), y -> [Int d; PP_tilda; print_month m; PP_tilda; Int y]

    let rec simpl = function
      |	None, y -> [<>]
      |	Some (None, m), y -> [<'None, y>]
      |	Some (Some d, m), y -> [<'Some (None, m), y>]

  end
