(** Intervals over values from the sub-logic. *)

module type PARAM =
  sig
    val verbose : bool
	(** When set to false the syntax of intervals is simply 'a..b'.
	   Otherwise it can be 'in a..b', '= a', '> a', '< a'. *)
  end

module Make (Param : PARAM) (Val : Logic.T) : Logic.T =
  struct
    include Logic.Default

    type bound = NegInf | PosInf | Val of Val.t

    let compare_bound a b = match a,b with
    | NegInf, NegInf
    | PosInf, PosInf -> 0
    | Val i, Val j -> Val.compare i j
    | a, b ->
	if a=NegInf or b=PosInf
	then -1
	else 1

    let max_of_min a b =
      let res = compare_bound a b in
      if res < 0 then b
      else if res > 0 then a
      else match a, b with
      |	Val x, Val y ->
	  if Val.entails x y then Val x
	  else if Val.entails y x then Val y
	  else failwith "Bug in the logic argument of Interv: compare x y = 0 ==> x |= y or y |= x."
      |	_ -> a
	
    let min_of_max a b =
      let res = compare_bound a b in
      if res < 0 then a
      else if res > 0 then b
      else match a, b with
      |	Val x, Val y ->
	  if Val.entails x y then Val x
	  else if Val.entails y x then Val y
	  else failwith "Bug in the logic argument of Interv: compare x y = 0 ==> x |= y or y |= x."
      |	_ -> a
	
    let max_of_max a b =
      let res = compare_bound a b in
      if res < 0 then b
      else if res > 0 then a
      else match a, b with
      |	Val x, Val y ->
	  if Val.entails x y then Val y
	  else if Val.entails y x then Val x
	  else failwith "Bug in the logic argument of Interv: compare x y = 0 ==> x |= y or y |= x."
      |	_ -> a
	
    let min_of_min a b =
      let res = compare_bound a b in
      if res < 0 then a
      else if res > 0 then b
      else match a, b with
      |	Val x, Val y ->
	  if Val.entails x y then Val y
	  else if Val.entails y x then Val x
	  else failwith "Bug in the logic argument of Interv: compare x y = 0 ==> x |= y or y |= x."
      |	_ -> a
	

    type t = bound * bound  (* invariant: in (a,b) one has (compare_bound a b <= 0) *)

    let isvalue = function
      | Val x, Val y -> Val.isvalue x & x=y
      | _ -> false

    let top () = (NegInf,PosInf)

    let entails_bound a b = match a, b with
    | Val x, Val y -> Val.entails x y
    | _ -> true

    let entails (a,b) (c,d) =
      let c1 = compare_bound c a in
      (c1 < 0 or (c1 = 0 & entails_bound a c)) &
      let c2 = compare_bound b d in
      (c2 < 0 or (c2 = 0 & entails_bound b d))

    let conj (a,b) (c,d) =
      let e = max_of_min a c in
      let f = min_of_max b d in
      if compare_bound e f <= 0 then (e,f) else raise Not_found

    let add f g = conj f g

    let sub f g = if entails f g then raise Not_found else f

    let features (a,b) =
      (true,(NegInf,PosInf)) ::
      match (a,b) with
      | Val x, Val y when x=y -> List.map (fun (vis,x) -> vis,(Val x,Val x)) (Val.features x)
      | _ -> [true,(a,b)]

    let gen (f1,f2) (g1,g2) hs =
      let x1, x2 = min_of_min f1 g1, max_of_max f2 g2 in
      if List.exists (fun h -> entails h (x1,x2)) hs
      then LSet.empty ()
      else LSet.of_list [(*(x1,PosInf);(NegInf,x2);*)(x1,x2)]

    let parse_interval =
      parser
	| [<a = Val.parse; s>] -> begin match s with parser
	  | [<'Token.DotDot>] -> begin match s with parser
	    | [<b = Val.parse>] ->
		if Val.compare a b <= 0 then (Val a,Val b)
		else raise (Stream.Error "Syntax error: inconsistent interval")
	    | [<>] -> (Val a,PosInf) end
	  | [<>] -> (Val a,Val a) end
	| [<'Token.DotDot; s>] -> begin match s with parser
	  | [<b = Val.parse>] -> (NegInf,Val b)
	  | [<>] -> (NegInf,PosInf) end
	| [<'Token.Ident "min">] -> (NegInf,NegInf)
	| [<'Token.Ident "max">] -> (PosInf,PosInf)

    let parse_verbose = parser
      |	[<'Token.Equal; str>] -> (match str with parser
           | [<x = Val.parse>] -> (Val x,Val x)
           | [<'Token.Ident "max">] -> (PosInf,PosInf)
           | [<'Token.Ident "min">] -> (NegInf,NegInf)
           | [<>] -> raise (Stream.Error "Syntax error: value expected after '='"))
      |	[<'Token.LT; x = Val.parse ?? "Syntax error: value expected after '<'">] -> (NegInf,Val x)
      |	[<'Token.GT; x = Val.parse ?? "Syntax error: value expected after '>'">] -> (Val x, PosInf)
      |	[<'Token.Ident "in"; a,b = parse_interval ?? "Syntax error: value interval expected after 'in'">] -> (a,b)

    let parse str =
      if Param.verbose
      then parse_verbose str
      else parse_interval str

    let print_interval = function
      | (NegInf,PosInf) -> [Token.DotDot]
      | (NegInf,NegInf) -> [Token.Ident "min"] 
      | (PosInf,PosInf) -> [Token.Ident "max"] 
      | (NegInf,Val b) -> Token.DotDot :: Token.PP_tilda :: Val.print b
      | (Val a,PosInf) -> Val.print a @ [Token.PP_tilda; Token.DotDot]
      | (Val a,Val b) ->
	  if a = b then Val.print a
	  else Val.print a @ [Token.PP_tilda; Token.DotDot; Token.PP_tilda] @ Val.print b
      | _ -> failwith "Bug in Interval.print: invalid interval"

    let print_verbose = function
      | (NegInf,PosInf) -> [Token.Ident "in"; Token.PP_tilda; Token.DotDot]
      | (NegInf,NegInf) -> [Token.Equal; Token.PP_tilda; Token.Ident "min"] 
      | (PosInf,PosInf) -> [Token.Equal; Token.PP_tilda; Token.Ident "max"] 
      | (NegInf,Val b) -> Token.LT :: Token.PP_tilda :: Val.print b
      | (Val a,PosInf) -> Token.GT :: Token.PP_tilda :: Val.print a
      | (Val a,Val b) ->
	  if a = b then Token.Equal :: Token.PP_tilda :: Val.print a
	  else Token.Ident "in" :: Token.PP_tilda :: Val.print a @ [Token.PP_tilda; Token.DotDot; Token.PP_tilda] @ Val.print b
      | _ -> failwith "Bug in Interval.print: invalid interval"

    let print f =
      if Param.verbose
      then print_verbose f
      else print_interval f

    let simpl = function
      | (NegInf,Val b) ->
          [<'(Val b,Val b); Common.stream_map (fun b' -> (NegInf,Val b')) (Val.simpl b)>]
      | (Val a,PosInf) ->
          [<'(Val a,Val a); Common.stream_map (fun a' -> (Val a',PosInf)) (Val.simpl a)>]
      | (Val a,Val b) ->
          if a=b
          then
            Common.stream_map (fun a' -> (Val a',Val a')) (Val.simpl a)
          else
            [<'(Val a,Val a); '(Val b,Val b);
              '(Val a,PosInf); '(NegInf,Val b);
              Common.stream_map (fun a' -> (Val a',Val b)) (Val.simpl a);
              Common.stream_map (fun b' -> (Val a,Val b')) (Val.simpl b)>]
      | _ -> [<>]

  end
