(** Logic signatures and common functions.

Logics are here understood as lattices of formulas, whose partial order
is called subsumption. Beyond lattice operations, other operations can
be required by applications.

*)

(** {1 Logic signatures} *)

(** Basic logical operations *)
module type T_BASE =
  sig
    type t (** Type of formulas. *)
    val parse : Syntax.t_stream -> t  (** Parser of formulas. *)
    val print : t -> Syntax.t_list (** Printer of formulas. *)
    val entails : t -> t -> bool (** Logical subsumption. *)
  end

(** Additional logical operations for CAMELIS *)
module type T_CAMELIS =
  sig
    include T_BASE
    val isvalue : t -> bool (** [isvalue f] returns whether [f] has neither more specific formula, nor different equivalents. *)
    val top : unit -> t (** Top formula. @raise Not_found when undefined *)
    val bot : unit -> t (** Bottom formula. @raise Not_found when undefined *)
    val conj : t -> t -> t (** Conjunction (glb). @raise Not_found when undefined *)
    val add : t -> t -> t (** Update the first formula by making it entail the second formula. @raise Not_found when undefined *)
    val sub : t -> t -> t (** Update the first formula by making it not entail the second formula. @raise Not_found when undefined *)
    val features : t -> (bool * t) list (** Extract a set of features entailed by the given formula. *)
	(** If the boolean is true, the feature is made visible. *)
    val axiom : t -> t option -> string LSet.t (** Get modified terms from axioms: snd arg. is None => Top. *)
    val gen : t -> t -> t list -> t LSet.t (** Dichotomic operator (for machine learning). *)
  end

(** Additional logical operations for DELFIN *)
module type T_DELFIN =
  sig
    include T_BASE
    val top : unit -> t (** Top formula. @raise Not_found when undefined *)
    val gen : t -> t -> t list -> t LSet.t (** Dichotomic operator (for machine learning). *)
    val simpl : t -> t Stream.t (** Produces a stream of syntactical simplifications of the given formula. *)
  end

(** The largest useful signature of logics. *)
module type T =
  sig
    type t
    val isvalue : t -> bool
    val top : unit -> t (* may raise Not_found *)
    val bot : unit -> t (* may raise Not_found *)
    val entails : t -> t -> bool
    val conj : t -> t -> t (* may raise Not_found *)
    val add : t -> t -> t (* may raise Not_found *)
    val sub : t -> t -> t (* may raise Not_found *)
    val features : t -> (bool * t) list
    val axiom : t -> t option -> string LSet.t
    val gen : t -> t -> t list -> t LSet.t
    val parse : Syntax.t_stream -> t
    val print : t -> Syntax.t_list
    val simpl : t -> t Stream.t
    val compare : t -> t -> int
    val parse_compact : Syntax.t_stream -> Syntax.t_list
  end

(** A default definition of a logic to be included in logic functors for extensibility. *)
module Default =
  struct
    let isvalue f = false
    let top () = raise Not_found
    let bot () = raise Not_found
    let entails f g = false
    let conj f g = raise Not_found
    let add f g = raise Not_found
    let sub f g = raise Not_found
    let features f = []
    let axiom f g_opt = LSet.empty ()
    let gen f g hs = LSet.empty ()
    let parse _ = raise (Invalid_argument "Logic.Default.parse: define 'parse' in logics")
    let print f = raise (Invalid_argument "Logic.Default.print: define 'print' in logics")
    let simpl f = [<>]
    let compare f g = 0
    let parse_compact _ = raise (Invalid_argument "Logic.Default.parse_compact: define 'parse_compact' in logics")
  end


(** {1 Common functions for logics} *)


let rec satgen =
  fun entails gen f g lazy_hs lubs ->
    let xs = gen f g lubs in
    if xs = []
    then []
    else
      let hs = Lazy.force lazy_hs in
      let xs' = List.filter (fun x -> not (List.exists (fun h -> entails h x) hs)) xs in
      if xs' = []
      then satgen entails gen f g lazy_hs (satgen_insert entails xs lubs)
      else xs'
and satgen_insert entails xs lubs =
  List.fold_left
    (fun res x -> x::List.filter (fun l -> not (entails x l)) lubs)
    lubs
    xs
