(** Union of the domains and languages of 2 sub-logics. *)

module Make (A1 : Logic.T) (A2 : Logic.T) : Logic.T =
  struct
    include Logic.Default

    type t = A1 of A1.t | A2 of A2.t

    let isvalue = function
      | A1 x -> A1.isvalue x
      | A2 x -> A2.isvalue x

    let top () = raise Not_found

    let bot () =
      try A1 (A1.bot ())
      with Not_found -> A2 (A2.bot ())

    let entails f g =
      match (f,g) with
      | A1 x, A1 y -> A1.entails x y
      | A2 x, A2 y -> A2.entails x y
      | _, _ -> false

    let conj f g =
      match (f,g) with
      | A1 x, A1 y -> A1 (A1.conj x y)
      | A2 x, A2 y -> A2 (A2.conj x y)
      | _, _ -> raise Not_found

    let add f g = conj f g

    let sub f g = if entails f g then raise Not_found else f

    let features = function
      | A1 x -> List.map (fun (vis,x) -> vis,A1 x) (A1.features x)
      | A2 x -> List.map (fun (vis,x) -> vis,A2 x) (A2.features x)

    let axiom f g_opt =
      match (f,g_opt) with
      | A1 x, None -> A1.axiom x None
      | A1 x, Some (A1 y) -> A1.axiom x (Some y)
      | A2 x, None -> A2.axiom x None
      | A2 x, Some (A2 y) -> A2.axiom x (Some y)
      | _, _ -> LSet.empty ()

    let gen f g hs =
      match (f,g) with
      | A1 x, A1 y ->
         List.map (fun t -> A1 t) (A1.gen x y (List.fold_right (fun h zs -> match (A1 x,h) with A1 x, A1 z -> z::zs | _, _ -> zs) hs []))
      | A2 x, A2 y ->
         List.map (fun t -> A2 t) (A2.gen x y (List.fold_right (fun h zs -> match (A2 x,h) with A2 x, A2 z -> z::zs | _, _ -> zs) hs []))
      | _, _ -> []

    let parse = parser
      | [<x = A1.parse>] -> A1 x
      | [<x = A2.parse>] -> A2 x

    let print = function
      | A1 x -> A1.print x
      | A2 x -> A2.print x

    let simpl = function
      | A1 x -> Common.stream_map (fun x -> A1 x) (A1.simpl x)
      | A2 x -> Common.stream_map (fun x -> A2 x) (A2.simpl x)

  end
