(** Times with 3 levels of granularity: hours, minutes, and seconds. *)

module Make : Logic.T =
  struct
    include Logic.Default

    type t = (int option * int) option * int  (* seconds, minutes, hours *)

    open Unix

    let now : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (Some (Some tm.tm_sec, tm.tm_min), tm.tm_hour)

    let this_min : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (Some (None, tm.tm_min), tm.tm_hour)

    let this_hour : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, tm.tm_hour)
    let next_hour : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, (tm.tm_hour + 1) mod 24)
    let last_hour : unit -> t = fun () ->
      let tm = localtime (time ()) in
      (None, (tm.tm_hour - 1) mod 24)


    let isvalue = function
      |	Some (Some s, m), h -> true
      | _ -> false

    let compare (sm1,h1) (sm2,h2) =
      if h1 < h2 then -1
      else if h1 > h2 then 1
      else match sm1, sm2 with
      |	None, _
      |	_, None -> 0
      |	Some (s1,m1), Some (s2,m2) ->
	  if m1 < m2 then -1
	  else if m1 > m2 then 1
	  else match s1, s2 with
	  | None, _
	  | _, None -> 0
	  | Some n1, Some n2 ->
	      if n1 < n2 then -1
	      else if n1 > n2 then 1
	      else 0

    let entails (sm1,h1) (sm2,h2) =
      h1 = h2 &
      match sm1, sm2 with
      |	_, None -> true
      |	Some (s1,m1), Some (s2,m2) ->
	  m1 = m2 &
	  ( match s1, s2 with
	  | _, None -> true
	  | Some n1, Some n2 -> n1 = n2
	  | _ -> false)
      |	_ -> false

    let conj t1 t2 =
      if entails t1 t2 then t1
      else if entails t2 t1 then t2
      else raise Not_found

    let add t1 t2 = conj t1 t2

    let sub t1 t2 = if entails t1 t2 then raise Not_found else t1

    let features = function
      |	None, h -> [true,(None, h)]
      |	Some (None, m), h -> [true,(Some (None, m), h); true,(None,h)]
      |	Some (Some s, m), h -> [true,(Some (Some s, m), h); true,(Some (None, m), h); true,(None,h)]


    open Token

    let parse = parser
      |	[<'Int h; ms>] ->
	  if h>=24 or h<0 then raise (Stream.Error ("Syntax error: out of range hour (0..23), in: " ^ Syntax.stringizer [Int h]));
	  ( match ms with parser
	  | [<'Colon; 'Int m ?? "Syntax error: minutes expected after: " ^ Syntax.stringizer [Int h; Colon]; s>] ->
	      if m>=60 or m<0 then raise (Stream.Error ("Syntax error: out of range minutes (0..59), in: " ^ Syntax.stringizer [Int h; Colon; Int m]));
	      ( match s with parser
	      |	[<'Colon; 'Int s ?? "Syntax error: seconds expected after: " ^ Syntax.stringizer [Int h; Colon; Int m; Colon]>] ->
		  if s>=60 or s<0 then raise (Stream.Error ("Syntax error: out of range seconds (0..59), in: " ^ Syntax.stringizer [Int h; Colon; Int m; Colon; Int s]));
		  (Some (Some s, m), h)
	      |	[<>] ->
		  (Some (None, m), h))
	  | [<'Ident "h">] ->
	      (None, h)
	  | [<>] ->
	      (None, h))
      |	[<'Ident "noon">] -> (Some (Some 0, 0), 12)
      |	[<'Ident "midnight">] -> (Some (Some 0, 0), 0)
      |	[<'Ident "now">] -> now ()
      |	[<'Ident "this"; str>] -> begin match str with parser
	| [<'Ident "sec">] -> now ()
	| [<'Ident "min">] -> this_min ()
	| [<'Ident "hour">] -> this_hour ()
	| [<>] -> raise (Stream.Error "Syntax error: 'sec', 'min' or 'hour' expected after 'this'") end
      |	[<'Ident "next"; 'Ident "hour" ?? "Syntax error: 'hour' expected after 'next'">] -> next_hour ()
      |	[<'Ident "last"; 'Ident "hour" ?? "Syntax error: 'hour' expected after 'last'">] -> last_hour ()

    let rec print = function
      |	None, h -> [Int h]
      |	Some (None, m), h -> [Int h; Colon; Int m]
      |	Some (Some s, m), h -> [Int h; Colon; Int m; Colon; Int s]

    let simpl = function
      | None, h -> [<>]
      | Some (None, m), h -> [<'None,h>]
      | Some (Some s, m), h -> [<'Some (None, m), h>]

  end 
